from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from .templates import format_chat
from typing import Union, Dict, List, Tuple

class TextTokenizer:
    """
    Applies template and tokenizes texts in a dataset.
    """

    __slots__ = ("tokenizer", "max_length", "prefix", "suffix", "gaps")

    def __init__(
        self,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
        template: str = "[CLS]{{text}}[SEP]",
        max_length: Union[int, None] = None,
    ) -> None:
        """
        Initializes the dataset tokenizer.

        Args:
            tokenizer: pre-trained transformers tokenizer.
            template: jinja2 template containing `{{text}}` keyword.
            max_length(optional): maximum sequence length after truncation.
        """
        assert "{{text}}" in template, "template should contain `{{text}}` substring"
        self.tokenizer: Union[PreTrainedTokenizerFast, PreTrainedTokenizer] = tokenizer
        self.max_length: Union[int, None] = max_length
        template_parts = template.split("{{text}}")
        self.prefix: List[int] = tokenizer(template_parts[0], add_special_tokens=False)["input_ids"]  # type: ignore
        self.suffix: List[int] = tokenizer(template_parts[1], add_special_tokens=False)["input_ids"]  # type: ignore
        self.gaps: Tuple = (len(self.prefix), len(self.suffix))
        if self.max_length != None:
            self.max_length -= len(self.prefix) + len(self.suffix)
            assert self.max_length > 0, "Max length should be larger than prompt length"

    def __call__(self, texts: List[str]) -> Dict[str, List[List[int]]]:
        """
        Tokenizes a batch of text samples.

        Args:
            texts: list of strings.

        Returns:
            dict with input_ids and prompt_mask for all text in a batch.
        """
        text_ids = self.tokenizer(
            texts,
            add_special_tokens=False,
            truncation=self.max_length != None,
            max_length=self.max_length,
        )["input_ids"]
        input_ids = [self.prefix + x + self.suffix for x in text_ids]  # type: ignore
        prompt_mask = [
            [1] * self.gaps[0] + [0] * len(x) + [1] * self.gaps[1] for x in text_ids  # type: ignore
        ]
        return {"input_ids": input_ids, "prompt_mask": prompt_mask}


class ChatTokenizer:
    """
    Applies chat template and tokenizes messages in a dataset.
    """

    __slots__ = ("tokenizer", "max_length")

    def __init__(
        self,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
        max_length: Union[int, None] = None,
    ) -> None:
        """
        Initializes the dataset chat tokenizer.

        Args:
            tokenizer: pre-trained transformers tokenizer.
            max_length(optional): maximum sequence length after truncation.
        """
        self.tokenizer: Union[PreTrainedTokenizerFast, PreTrainedTokenizer] = tokenizer
        self.max_length: Union[int, None] = max_length

    def __call__(
        self,
        prompts: Union[List[str], List[List[str]]],
        responses: Union[List[str], List[List[str]]] = [],
        system: Union[List[str], None] = None,
    ) -> Dict[str, List[List[int]]]:
        """
        Tokenizes chats.

        Args:
            prompts: a list where each item is either a single prompt or multiple prompts in a dialog.
            responses: a list where each item is either a single response or multiple responses in a dialog.

        Returns:
            dict with input_ids and prompt_mask for all text in a batch.
        """
        batch_size = len(prompts)
        chats = [
            format_chat(prompts[i], responses[i], system[i] if system else None)
            for i in range(batch_size)
        ]
        inputs = self.tokenizer.apply_chat_template(
            chats,
            return_assistant_tokens_mask=True,
            tokenize=True,
            return_dict=True,
            truncation=self.max_length != None,
            max_length=self.max_length,
        )
        prompt_mask = [
            [0 if x == 1 else 1 for x in tokens] for tokens in inputs["assistant_masks"]  # type: ignore
        ]
        input_ids = inputs["input_ids"]  # type: ignore
        return {"input_ids": input_ids, "prompt_mask": prompt_mask}  # type: ignore



import os
from datasets import Dataset, load_from_disk
from typing import Tuple, Callable, List
from torch.utils.data import DataLoader


def split_dataset(
    dataset: Dataset, train_size: float = 0.8, val_size: float = 0.1, seed: int = 42
) -> Tuple[Dataset, Dataset, Dataset]:
    """
    Split HF dataset into train, validation and test.

    Args:
        dataset: HF dataset.
        train_size: train ratio from 0 to 1.
        val_size: validation ratio from 0 to 1.
        seed: seed used for splitting.

    Returns:
        a tuple of 3 datasets for train, validation and test.

    train_size and val_size are taken from total and their sum should not be more than 1.

    Test size would be equal to 1 - train_size - val_size.
    """
    split1 = dataset.train_test_split(train_size=train_size, seed=seed)  # type: ignore
    train = split1["train"]
    temp = split1["test"]
    temp_val_size = val_size / (1.0 - train_size)
    split2 = temp.train_test_split(train_size=temp_val_size, seed=seed)  # type: ignore
    val = split2["train"]
    test = split2["test"]
    return train, val, test


def load_cached(cache_path: str, process_fn: Callable) -> Dataset:
    """
    Try loading processed dataset from cache. Processes dataset and saves it to cache if pre-cached dataset is not found.

    Args:
        cache_path: path to load/save dataset.
        process_fn: function that will return a new processed dataset if cache is not found.

    Returns:
        a pre-processed HF dataset.
    """
    if os.path.exists(cache_path):
        print(f"Loading dataset from '{cache_path}'...")
        ds = load_from_disk(cache_path)
    else:
        ds = process_fn()
        print("Saving processed dataset to disk...")
        ds.save_to_disk(cache_path)
        print(f"Processed dataset saved to '{cache_path}'")
    print("Dataset is ready!")
    return ds  # type: ignore


def get_loaders(
    dataset: Dataset,
    collate_fn: Callable = lambda x: x,
    batch_size: int = 16,
    train_size: float = 0.8,
    val_size: float = 0.1,
    num_workers: int = 15,
    seed: int = 42,
) -> Tuple[DataLoader, DataLoader, DataLoader]:
    """
    Shortcut to generate pytorch dataloaders (train/val/test) from hf dataset.

    Args:
        dataset: HF dataset.
        collate_fn: function used for dataset collation.
        batch_size: batch_size for dataloaders.
        train_size: train split size. see `split_dataset`.
        val_size: validation split size. see `split_dataset`.
        num_workers: number of pytorch dataloader workers.
        seed: random seed used for splitting.

    Returns:
        a tuple with train, validation and test pytorch dataloaders.

    Splits dataset and assigns collators automatically.
    """
    train, val, test = split_dataset(
        dataset, train_size=train_size, val_size=val_size, seed=seed
    )
    train_loader = DataLoader(train, batch_size=batch_size, collate_fn=collate_fn, num_workers=num_workers, shuffle=True)  # type: ignore
    val_loader = DataLoader(val, batch_size=batch_size, collate_fn=collate_fn, num_workers=num_workers, shuffle=False)  # type: ignore
    test_loader = DataLoader(test, batch_size=batch_size, collate_fn=collate_fn, num_workers=num_workers, shuffle=False)  # type: ignore
    return train_loader, val_loader, test_loader


def strip_features(
    dataset: Dataset, keep_features: List[str] = ["input_ids", "prompt_mask"]
) -> Dataset:
    """
    Remove all features from dataset except specified ones.

    Args:
        dataset: HF dataset to purify.
        keep_features: list of feature names to keep.

    Returns:
        a new dataset with only specified features.

    Useful for reducing memory usage and cleaning up datasets.
    """

    # Get current features
    current_features = set(dataset.features.keys())
    features_to_remove = current_features - set(keep_features)

    if not features_to_remove:
        print(f"Dataset already has only {keep_features} features")
        return dataset

    print(f"Removing features: {list(features_to_remove)}")
    print(f"Keeping features: {keep_features}")

    # Remove unwanted features
    return dataset.remove_columns(list(features_to_remove))



from typing import Union, Callable, Tuple
from torch import batch_norm
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from datasets import load_dataset, concatenate_datasets
from torch.utils.data import DataLoader
from .processors import ChatTokenizer, TextTokenizer
from .filters import LengthFilter
from .utils import load_cached, get_loaders, strip_features
from .collators import PadCollator

"""
Pre-processing functions for some standard datasets
"""


def build_dataset(preprocess_func):
    """
    Decorator for building dataloaders. Processes or tries loading the dataset if `cache_path` is specified, creates new PadCollator, and returns dataloaders.
    """

    def wrapper(*args, **kwargs):
        cache_path = kwargs.get("cache_path", None)
        # Load from cache
        if cache_path != None:
            ds = load_cached(cache_path, lambda: preprocess_func(*args, **kwargs))
        else:
            ds = preprocess_func(*args, **kwargs)

        # Split and create dataloaders
        tokenizer = kwargs.get("tokenizer", None)
        tokenizer = args[0] if tokenizer == None and len(args) > 0 else None
        assert tokenizer != None
        pre = kwargs.get("pre", None)
        post = kwargs.get("post", None)
        collate_fn = PadCollator(pad_id=tokenizer.pad_token_id, pre=pre, post=post)  # type: ignore

        # Return dataloaders
        batch_size = kwargs.get("batch_size", 16)
        seed = kwargs.get("seed", 42)
        return get_loaders(dataset=ds, collate_fn=collate_fn, batch_size=batch_size, seed=seed)  # type: ignore

    return wrapper


@build_dataset
def alpaca(
    tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
    cache_path: Union[str, None] = None,
    max_length: int = 512,
    batch_size: int = 16,
    seed: int = 42,
    pre: Union[Callable, None] = None,
    post: Union[Callable, None] = None,
) -> Tuple[DataLoader, DataLoader, DataLoader]:
    """
    Downloads and Pre-processes [Alpaca](https://huggingface.co/datasets/tatsu-lab/alpaca) dataset for instruction fine-tuning.

    Args:
        tokenizer: transformers tokenizer with `chat_template` parameter set.
        cache_path: path to load/save dataset.
        max_length: maximum sequence length for filter.
        batch_size: dataloaders batch size.
        seed: splitting and shuffling seed.
        pre: see `PadCollator`.
        post: see `PadCollator`.

    Returns:
        a tuple of train, validation and test dataloaders.
    """

    dataset_name = "tatsu-lab/alpaca"
    print(f"Loading {dataset_name} dataset (train split)...")
    ds = load_dataset(dataset_name, split="train")
    print("Shuffling dataset...")
    ds = ds.shuffle(seed=seed)
    print("Processing prompts...")
    ds = ds.map(
        lambda instructions, inputs: {
            "prompt": [
                f"{instructions[i]}\n{inputs[i]}" if inputs[i] else instructions[i]
                for i in range(len(instructions))
            ]
        },
        input_columns=["instruction", "input"],
        batched=True,
        batch_size=512,
    )
    print("Tokenization...")
    ds = ds.map(
        ChatTokenizer(tokenizer),
        input_columns=["prompt", "output"],
        batched=True,
        batch_size=512,
    )
    print("Removing features...")
    ds = strip_features(ds)  # type: ignore
    print("Filtering...")
    old_size = len(ds)  # type: ignore
    ds = ds.filter(
        LengthFilter(min_length=0, max_length=max_length),
        input_columns=["input_ids"],
        batched=True,
        batch_size=512,
    )
    new_size = len(ds)  # type: ignore
    print(
        f"{old_size - new_size} samples were removed ({(1.0-new_size/old_size)*100:.2f}%)"
    )
    return ds # type: ignore


@build_dataset
def tinystories(
    tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
    template: str = "[CLS]{{tex}}[SEP]",
    cache_path: Union[str, None] = None,
    max_length: int = 512,
    batch_size: int = 16,
    seed: int = 42,
    pre: Union[Callable, None] = None,
    post: Union[Callable, None] = None,
) -> Tuple[DataLoader, DataLoader, DataLoader]:
    """
    Downloads and Pre-processes [TinyStories](https://huggingface.co/datasets/roneneldan/TinyStories) dataset for language modeling. `train` and `validation` are merged, shuffled and re-split with 80%/10%/10% ratio.

    Args:
        tokenizer: transformers tokenizer.
        template: prompt template with `{{text}}` placeholder.
        cache_path: path to load/save dataset.
        max_length: maximum sequence length for truncation.
        batch_size: dataloaders batch size.
        seed: splitting and shuffling seed.
        pre: see `PadCollator`.
        post: see `PadCollator`.

    Returns:
        a tuple of train, validation and test dataloaders.
    """

    dataset_name = "roneneldan/TinyStories"
    print(f"Loading {dataset_name} dataset (all splits)...")
    ds = load_dataset(dataset_name)
    print("Merging train and validation splits...")
    ds = concatenate_datasets([ds["train"], ds["validation"]])  # type:ignore
    print("Shuffling dataset...")
    ds = ds.shuffle(seed=seed)
    print("Tokenization...")
    ds = ds.map(
        TextTokenizer(tokenizer, template=template, max_length=max_length),
        input_columns=["text"],
        batched=True,
        batch_size=512,
    )
    print("Removing features...")
    ds = strip_features(ds)  # type: ignore
    print("Filtering...")
    old_size = len(ds)  # type: ignore
    ds = ds.filter(
        LengthFilter(min_length=0, max_length=max_length),
        input_columns=["input_ids"],
        batched=True,
        batch_size=512,
    )
    new_size = len(ds)  # type: ignore
    print(
        f"{old_size - new_size} samples were removed ({(1.0-new_size/old_size)*100:.2f}%)"
    )
    return ds # type: ignore



import torch
from torch.nn.utils.rnn import pad_sequence
from typing import Union, Dict, List, Callable, Literal, Any


class PadCollator:
    """
    Collator which converts inputs into pytorch tensors and applies padding.
    """

    __slots__ = ("pad_id", "pre", "post", "pad_fn", "ignore_columns")

    def __init__(
        self,
        pad_id: int,
        pre: Union[Callable, None] = None,
        post: Union[Callable, None] = None,
        padding_side: Literal["right", "left"] = "right",
        ignore_columns: List[str] = [],
    ) -> None:
        """
        Initializes the collator.

        Args:
            pad_id: pad token id used for padding.
            pre: callable that accepts list of dicts where every key is a dataset feature, modifies it and returns it back.
            post: callable that accepts a dict of tensors of a form `{"input_ids": ..., "attention_mask": ..., "prompt_mask": ...}`, modifies it and returns it back.
            padding_side: can be right or left.
            ignore_columns: columns to be excluded.
        """
        self.pad_fn: Callable = lambda batch, column, pad: pad_sequence(
            [torch.tensor(x[column]) for x in batch],
            batch_first=True,
            padding_value=pad,
            padding_side=padding_side,
        )
        self.pre = pre
        self.post = post
        self.pad_id: int = pad_id
        self.ignore_columns = ignore_columns

    def __call__(self, batch: List[Dict[str, Any]]) -> Dict[str, torch.Tensor]:
        """
        Collates inputs.

        Args:
            batch: list of dicts where each key is a feature.

        Returns:
            a dict of padded tensors of form `{"input_ids": ..., "attention_mask": ..., "prompt_mask": ...}`.
        """
        # Apply pre-processing
        if self.pre != None:
            batch = self.pre(batch)
        # Pad inputs
        # Use temp pad ids (-42) to prevent excluding intentional pad tokens
        input_ids = self.pad_fn(batch, "input_ids", -42)
        prompt_mask = self.pad_fn(batch, "prompt_mask", 1)
        # Generate attention_mask
        attention_mask = torch.ones_like(input_ids)
        attention_mask[input_ids == -42] = 0
        # Set proper pad ids
        input_ids[input_ids == -42] = self.pad_id
        # Apply post
        output = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "prompt_mask": prompt_mask,
        }
        # Add other columns
        processed_columns = output.keys()
        other_columns = batch[0].keys()
        for column in other_columns:
            if column in processed_columns or column in self.ignore_columns:
                continue
            output[column] = []
            for sample in batch:
                output[column].append(sample[column])
        # Apply post-processing
        if self.post != None:
            output = self.post(output)
        return output



"""
LOTD - Lord of the Datasets
Efficient NLP dataset preprocessing library.
"""

import importlib.metadata
from . import dataset_builders as datasets
from .collators import PadCollator
from .processors import TextTokenizer, ChatTokenizer
from .filters import LengthFilter
from .utils import split_dataset, load_cached, get_loaders, strip_features
from .templates import generate_chat_template, format_chat

try:
    __version__ = importlib.metadata.version("lotd")
except importlib.metadata.PackageNotFoundError:
    # Fallback for development/editable installs
    __version__ = "0.0.0"

__all__ = [
    "datasets",
    "PadCollator",
    "TextTokenizer",
    "ChatTokenizer",
    "LengthFilter",
    "split_dataset",
    "load_cached",
    "get_loaders",
    "strip_features",
    "generate_chat_template",
    "format_chat",
]



import sys
from typing import List, Union

class LengthFilter:
    """
    Filters datasets by max and min length of input_ids.
    """

    __slots__ = ("max_length", "min_length")

    def __init__(self, min_length: int = 0, max_length: Union[int, None] = None) -> None:
        """
        Initializes the collator.

        Args:
            min_length: lowest number of tokens.
            max_length: highest number of tokens.
        """
        self.max_length = sys.maxsize if max_length == None else max_length
        self.min_length = min_length


    def __call__(self, input_ids: List[List[int]]) -> List[bool]:
        """
        Processes a batch of input_ids.
        """
        return [len(x) <= self.max_length and len(x) > self.min_length for x in input_ids]




from typing import Union, Dict, List, Tuple


def format_chat(
    prompts: Union[str, List[str]] = [],
    responses: Union[str, List[str]] = [],
    system: Union[str, None] = None,
) -> List[Dict[str, str]]:
    """
    Format list of prompts, responses and system messages into a chat.

    Args:
        prompts: a single prompt or a list of prompts provided by a user in a dialog.
        responses: a single response or a list of responses producrd by a model in a dialog.
        system: system prompt used for this chat.

    Returns:
        a list of messages of form `[{"role": "...", "content": "..."}, ...]`.
    """
    # Convert to unified form
    if isinstance(prompts, str):
        prompts = [prompts]
    if isinstance(responses, str):
        responses = [responses]
    # Check input data
    assert len(prompts) == len(responses), "each prompt should have a response"
    # Convert to chat array
    chat = []
    if system:
        chat.append({"role": "system", "content": system})
    for i in range(len(prompts)):
        chat.append({"role": "user", "content": prompts[i]})
        chat.append({"role": "assistant", "content": responses[i]})
    return chat


def generate_chat_template(
    last_only: bool = False,
    wrapper: str = "{{template}}",
    system_tags: Tuple[str, str] = ("<|system|>", "<|end|>"),
    user_tags: Tuple[str, str] = ("<|user|>", "<|end|>"),
    assistant_tags: Tuple[str, str] = ("<|assistant|>", "<|end|>"),
) -> str:
    """
    Generate chat template for instruction tuning.

    Args:
        last_only: only enable generation for last response. If true, assistant_masks ignores previous responses. Default: False.
        wrapper: a wrapper for adding text before and after the template. Should contain `{{template}}`.
        system_tags: a pair of tags to mark the beginning and the end of a system role in a dialog.
        user_tags: see `system_tags`.
        assistant_tags: see `system_tags`.

    Returns:
        a jinja2 template string.
    """
    assert "{{template}}" in wrapper, "wrapper should contain `{{template}}` substring"
    template = """
{% for message in messages %}
{% if message['role'] == 'system' -%}
system_start{{message['content']}}system_end
{% elif message['role'] == 'user' -%}
user_start{{message['content']}}user_end
{% elif message['role'] == 'assistant' -%}
assistant_message
{% endif -%}
{% endfor -%}
""".strip()
    # Insert correct assistant message template
    if last_only:
        template = template.replace(
            "assistant_message",
            "assistant_start{% if loop.last %}{% generation %}{{message['content']}}{% endgeneration %}{% else %}{{message['content']}}{% endif %}assistant_end",
        )
    else:
        template = template.replace(
            "assistant_message",
            "assistant_start{% generation %}{{message['content']}}{% endgeneration %}assistant_end",
        )
    # Insert message tags
    for i, x in enumerate(["start", "end"]):
        template = template.replace(f"system_{x}", system_tags[i])
        template = template.replace(f"user_{x}", user_tags[i])
        template = template.replace(f"assistant_{x}", assistant_tags[i])
    # Wrap template
    template = wrapper.replace("{{template}}", template)
    return template
